# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python code parsing."""

import ast
import inspect
import io
import re

import langfun.core as lf
from langfun.core.coding.python import errors
from langfun.core.coding.python import permissions


class PythonCodeParser(lf.Component):
  """Python code parser with permission control."""

  _ID_REGEX = re.compile('^[a-zA-Z_\\-]*$')

  class _CodeValidator(ast.NodeVisitor):
    """Python AST node visitor for ensuring code are permitted."""

    def __init__(self, code: str, permission: permissions.CodePermission):
      super().__init__()
      self.code = code
      self.permission = permission

    def verify(
        self,
        node,
        flag: permissions.CodePermission,
        node_type,
        error_message: str,
    ) -> None:
      if isinstance(node, node_type) and not (self.permission & flag):
        raise SyntaxError(
            error_message, (
                '<generated-code>',
                node.lineno,
                node.col_offset,
                self._code_line(node.lineno),
                node.end_lineno,
                node.end_col_offset,
            ))

    def _code_line(self, lineno):
      return self.code.split('\n')[lineno - 1]

    def generic_visit(self, node):
      self.verify(
          node,
          permissions.CodePermission.CONDITION,
          (ast.If, ast.Match),
          'Condition is not allowed.',
      )

      self.verify(
          node,
          permissions.CodePermission.LOOP,
          (ast.For, ast.While, ast.AsyncFor, ast.AsyncWith),
          'Loop is not allowed.',
      )

      self.verify(
          node,
          permissions.CodePermission.EXCEPTION,
          (ast.Try, ast.Raise, ast.Assert),
          'Exception is not allowed.',
      )

      self.verify(
          node,
          permissions.CodePermission.CLASS_DEFINITION,
          ast.ClassDef,
          'Class definition is not allowed.',
      )

      self.verify(
          node,
          permissions.CodePermission.FUNCTION_DEFINITION,
          (
              ast.FunctionDef,
              ast.AsyncFunctionDef,
              ast.Return,
              ast.Yield,
              ast.YieldFrom,
          ),
          'Function definition is not allowed.',
      )

      self.verify(
          node,
          permissions.CodePermission.IMPORT,
          (ast.Import, ast.ImportFrom),
          '`import` is not allowed.',
      )

      super().generic_visit(node)

  def parse(
      self, code: str, permission: permissions.CodePermission
  ) -> tuple[str, ast.AST]:
    code = self.clean(code)
    try:
      parsed_code = ast.parse(code, mode='exec')
      PythonCodeParser._CodeValidator(code, permission).visit(parsed_code)
    except SyntaxError as e:
      raise errors.CodeError(code, e) from e
    return code, parsed_code

  def clean(self, code_text: str) -> str:
    # TODO(daiyip): Deal with markdown in docstrings.
    code = io.StringIO()
    quote_char = None
    in_code = False
    i = 0
    in_comment = False
    while i < len(code_text):
      c = code_text[i]
      # Detect code block separator (```).
      if (not in_comment
          and quote_char is None
          and c == '`'
          and code_text[i:i + 3] == '```'):
        in_code = not in_code
        if in_code:
          i += 3
          continue
        else:
          break

      # Detect string literal boundary.
      if (in_code
          and not in_comment
          and c in ('\'', '"')
          and i > 0
          and code_text[i - 1] != '\\'):
        # Handle ''' and """.
        if code_text[i: i + 3] == c * 3:
          c = c * 3
          i += 2

        if quote_char is None:
          quote_char = c
        elif quote_char == c:
          # NOTE(daiyip): at times, LM forgets to escape quotes inside a string.
          # Thus we do some smart checking here to automatically correct such
          # case. This logic here is pretty involved in handling special cases.
          # We might want to revisit them later.

          # Peek forward to see if it could be a valid string.
          nt, nnt_start = _next_token(code_text, i + 1)
          if (len(c) == 3
              or nt in (',', '[', ']', '}', ')', '+', '*', '%', '\n', ':')):
            end_quote = True
          elif nt == ' ':
            # Detect if . could be a method invocation.
            # NOTE(daiyip): 'in' and 'not in' might have false positives. But
            # given the chance is low, we do not complicate the reasoning logic
            # for now.
            nnt, _ = _next_token(code_text, nnt_start, skip_whitespace=True)
            end_quote = nnt in ('+', '*', '%', '#', '[', 'in', 'not', ':')
          elif nt == '.':
            # Detect if . could be method invocation on string.
            nnt, nnnt_start = _next_token(code_text, nnt_start)
            nnnt, _ = _next_token(code_text, nnnt_start)
            end_quote = nnt.isidentifier() and nnnt == '('
          else:
            end_quote = False

          if end_quote:
            quote_char = None
          else:
            c = f'\\{c}'
      # Detect comment.
      elif c == '#' and quote_char is None:
        in_comment = True
      # Detect end-of-comment.
      elif c == '\n':
        # NOTE(daiyip): deal with cases that LM forgot to escape linebreaks
        # within strings.
        if quote_char is not None:
          # Only add \\ for ' and " (other than ''' and """).
          if len(quote_char) == 1:
            c = '\\n'
        else:
          in_comment = False

      if in_code:
        code.write(c)

      i += 1

    code = code.getvalue()
    if code:
      pos = code.find('\n')
      # Strip markdown code type. E.g. ```python
      if pos > 0 and self._ID_REGEX.match(code[:pos]):
        code = code[pos:]
    else:
      # Maybe-code that resides not within a code markdown block.
      # Adding '\n' makes inspect.cleandoc to make right adjustment.
      code = '\n' + code_text
    return inspect.cleandoc(code).strip()


def _next_token(
    text: str,
    start: int = 0,
    skip_whitespace: bool = False
    ) -> tuple[str, int]:
  """Find the next token in a string with a start position."""
  token_start = start
  if skip_whitespace:
    while token_start < len(text) and text[token_start] in ' \t':
      token_start += 1
  token_end = token_start + 1
  if text[token_start].isalpha():
    while (token_end < len(text)
           and text[token_end].isalpha() or text[token_end] in '_'):
      token_end += 1
  return text[token_start:token_end], token_end



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for OpenAI models."""

import unittest
from unittest import mock

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import openai
import pyglove as pg


def mock_completion_query(prompt, *, n=1, **kwargs):
  del kwargs
  choices = []
  for i, _ in enumerate(prompt):
    for k in range(n):
      choices.append(pg.Dict(
          index=i,
          text=f'Sample {k} for prompt {i}.',
          logprobs=k / 10,
      ))
  return pg.Dict(
      choices=choices,
      usage=lf.LMSamplingUsage(
          prompt_tokens=100,
          completion_tokens=100,
          total_tokens=200,
      ),
  )


def mock_chat_completion_query(messages, *, n=1, **kwargs):
  if len(messages) > 1:
    system_message = f' system={messages[0]["content"]}'
  else:
    system_message = ''

  if 'response_format' in kwargs:
    response_format = f' format={kwargs["response_format"]["type"]}'
  else:
    response_format = ''

  choices = []
  for k in range(n):
    choices.append(pg.Dict(
        message=pg.Dict(
            content=(
                f'Sample {k} for message.{system_message}{response_format}'
            )
        ),
        logprobs=None,
    ))
  return pg.Dict(
      choices=choices,
      usage=lf.LMSamplingUsage(
          prompt_tokens=100,
          completion_tokens=100,
          total_tokens=200,
      ),
  )


def mock_chat_completion_query_vision(messages, *, n=1, **kwargs):
  del kwargs
  choices = []
  urls = [
      c['image_url']['url']
      for c in messages[0]['content'] if c['type'] == 'image_url'
  ]
  for k in range(n):
    choices.append(pg.Dict(
        message=pg.Dict(
            content=f'Sample {k} for message: {"".join(urls)}'
        ),
        logprobs=None,
    ))
  return pg.Dict(
      choices=choices,
      usage=lf.LMSamplingUsage(
          prompt_tokens=100,
          completion_tokens=100,
          total_tokens=200,
      ),
  )


class OpenAITest(unittest.TestCase):
  """Tests for OpenAI language model."""

  def test_model_id(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key').model_id, 'OpenAI(text-davinci-003)')

  def test_resource_id(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key').resource_id, 'OpenAI(text-davinci-003)'
    )

  def test_max_concurrency(self):
    self.assertGreater(openai.Gpt35(api_key='test_key').max_concurrency, 0)

  def test_get_request_args(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key', timeout=90.0)._get_request_args(
            lf.LMSamplingOptions(
                temperature=2.0,
                logprobs=True,
                n=2,
                max_tokens=4096,
                top_p=1.0)),
        dict(
            engine='text-davinci-003',
            logprobs=True,
            top_logprobs=None,
            n=2,
            temperature=2.0,
            max_tokens=4096,
            stream=False,
            timeout=90.0,
            top_p=1.0,
        )
    )
    self.assertEqual(
        openai.Gpt4(api_key='test_key')._get_request_args(
            lf.LMSamplingOptions(
                temperature=1.0, stop=['\n'], n=1, random_seed=123
            )
        ),
        dict(
            model='gpt-4',
            top_logprobs=None,
            n=1,
            temperature=1.0,
            stream=False,
            timeout=120.0,
            stop=['\n'],
            seed=123,
        ),
    )
    with self.assertRaisesRegex(RuntimeError, '`logprobs` is not supported.*'):
      openai.GptO1Preview(api_key='test_key')._get_request_args(
          lf.LMSamplingOptions(
              temperature=1.0, logprobs=True
          )
      )

  def test_call_completion(self):
    with mock.patch('openai.Completion.create') as mock_completion:
      mock_completion.side_effect = mock_completion_query
      lm = openai.OpenAI(api_key='test_key', model='text-davinci-003')
      self.assertEqual(
          lm('hello', sampling_options=lf.LMSamplingOptions(n=2)),
          'Sample 0 for prompt 0.',
      )

  def test_call_chat_completion(self):
    with mock.patch('openai.ChatCompletion.create') as mock_chat_completion:
      mock_chat_completion.side_effect = mock_chat_completion_query
      lm = openai.OpenAI(api_key='test_key', model='gpt-4')
      self.assertEqual(
          lm('hello', sampling_options=lf.LMSamplingOptions(n=2)),
          'Sample 0 for message.',
      )

  def test_call_chat_completion_vision(self):
    with mock.patch('openai.ChatCompletion.create') as mock_chat_completion:
      mock_chat_completion.side_effect = mock_chat_completion_query_vision
      lm_1 = openai.Gpt4Turbo(api_key='test_key')
      lm_2 = openai.Gpt4VisionPreview(api_key='test_key')
      for lm in (lm_1, lm_2):
        self.assertEqual(
            lm(
                lf.UserMessage(
                    'hello <<[[image]]>>',
                    image=lf_modalities.Image.from_uri('https://fake/image')
                ),
                sampling_options=lf.LMSamplingOptions(n=2)
            ),
            'Sample 0 for message: https://fake/image',
        )

  def test_sample_completion(self):
    with mock.patch('openai.Completion.create') as mock_completion:
      mock_completion.side_effect = mock_completion_query
      lm = openai.OpenAI(api_key='test_key', model='text-davinci-003')
      results = lm.sample(
          ['hello', 'bye'], sampling_options=lf.LMSamplingOptions(n=3)
      )

    self.assertEqual(len(results), 2)
    self.assertEqual(
        results[0],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for prompt 0.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for prompt 0.',
                        score=0.1,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.1,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for prompt 0.',
                        score=0.2,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.2,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=50, completion_tokens=50, total_tokens=100
            ),
        ),
    )
    self.assertEqual(
        results[1],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for prompt 1.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for prompt 1.',
                        score=0.1,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.1,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for prompt 1.',
                        score=0.2,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=16,
                            completion_tokens=16,
                            total_tokens=33
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.2,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=50, completion_tokens=50, total_tokens=100
            ),
        ),
    )

  def test_sample_chat_completion(self):
    with mock.patch('openai.ChatCompletion.create') as mock_chat_completion:
      mock_chat_completion.side_effect = mock_chat_completion_query
      openai.SUPPORTED_MODELS_AND_SETTINGS['gpt-4'].update({
          'cost_per_1k_input_tokens': 1.0,
          'cost_per_1k_output_tokens': 1.0,
      })
      lm = openai.OpenAI(api_key='test_key', model='gpt-4')
      results = lm.sample(
          ['hello', 'bye'], sampling_options=lf.LMSamplingOptions(n=3)
      )

    self.assertEqual(len(results), 2)
    print(results[0])
    self.assertEqual(
        results[0],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200,
                estimated_cost=0.2,
            ),
        ),
    )
    self.assertEqual(
        results[1],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 2 for message.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=33,
                            completion_tokens=33,
                            total_tokens=66,
                            estimated_cost=0.2 / 3,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200,
                estimated_cost=0.2,
            ),
        ),
    )

  def test_sample_with_contextual_options(self):
    with mock.patch('openai.Completion.create') as mock_completion:
      mock_completion.side_effect = mock_completion_query
      lm = openai.OpenAI(api_key='test_key', model='text-davinci-003')
      with lf.use_settings(sampling_options=lf.LMSamplingOptions(n=2)):
        results = lm.sample(['hello'])

    self.assertEqual(len(results), 1)
    self.assertEqual(
        results[0],
        lf.LMSamplingResult(
            [
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 0 for prompt 0.',
                        score=0.0,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=50,
                            completion_tokens=50,
                            total_tokens=100,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.0,
                    logprobs=None,
                ),
                lf.LMSample(
                    lf.AIMessage(
                        'Sample 1 for prompt 0.',
                        score=0.1,
                        logprobs=None,
                        is_cached=False,
                        usage=lf.LMSamplingUsage(
                            prompt_tokens=50,
                            completion_tokens=50,
                            total_tokens=100,
                        ),
                        tags=[lf.Message.TAG_LM_RESPONSE],
                    ),
                    score=0.1,
                    logprobs=None,
                ),
            ],
            usage=lf.LMSamplingUsage(
                prompt_tokens=100, completion_tokens=100, total_tokens=200
            ),
        ),
    )

  def test_call_with_system_message(self):
    with mock.patch('openai.ChatCompletion.create') as mock_chat_completion:
      mock_chat_completion.side_effect = mock_chat_completion_query
      lm = openai.OpenAI(api_key='test_key', model='gpt-4')
      self.assertEqual(
          lm(
              lf.UserMessage(
                  'hello',
                  system_message='hi',
              ),
              sampling_options=lf.LMSamplingOptions(n=2)
          ),
          'Sample 0 for message. system=hi',
      )

  def test_call_with_json_schema(self):
    with mock.patch('openai.ChatCompletion.create') as mock_chat_completion:
      mock_chat_completion.side_effect = mock_chat_completion_query
      lm = openai.OpenAI(api_key='test_key', model='gpt-4')
      self.assertEqual(
          lm(
              lf.UserMessage(
                  'hello',
                  json_schema={
                      'type': 'object',
                      'properties': {
                          'name': {'type': 'string'},
                      },
                      'required': ['name'],
                      'title': 'Person',
                  }
              ),
              sampling_options=lf.LMSamplingOptions(n=2)
          ),
          'Sample 0 for message. format=json_schema',
      )

    # Test bad json schema.
    with self.assertRaisesRegex(ValueError, '`json_schema` must be a dict'):
      lm(lf.UserMessage('hello', json_schema='foo'))

    with self.assertRaisesRegex(
        ValueError, 'The root of `json_schema` must have a `title` field'
    ):
      lm(lf.UserMessage('hello', json_schema={}))


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for text formatting."""

import inspect
import unittest
from langfun.core import text_formatting


class TextFormattingTest(unittest.TestCase):

  def test_colored_template(self):
    original_text = inspect.cleandoc("""
        Hi {{ foo }}
        {# print x if x is present #}
        {% if x %}
        {{ x }}
        {% endif %}
        """)

    colored_text = text_formatting.colored_template(
        text_formatting.colored(original_text, color='blue')
    )
    self.assertEqual(
        colored_text,
        '\x1b[34mHi \x1b[44m\x1b[37m{{ foo }}\x1b[0m\x1b[34m\n'
        '\x1b[32m{# print x if x is present #}\x1b[0m\x1b[34m\n'
        '\x1b[31m{% if x %}\x1b[0m\x1b[34m\n'
        '\x1b[44m\x1b[37m{{ x }}\x1b[0m\x1b[34m\n'
        '\x1b[31m{% endif %}\x1b[0m\x1b[34m\x1b[0m'
    )
    self.assertEqual(text_formatting.decolored(colored_text), original_text)

  def test_colored_without_termcolor(self):
    termcolor = text_formatting.termcolor
    text_formatting.termcolor = None
    original_text = inspect.cleandoc("""
        Hi {{ foo }}
        {# print x if x is present #}
        {% if x %}
        {{ x }}
        {% endif %}
        """)

    colored_text = text_formatting.colored_template(
        text_formatting.colored(original_text, color='blue')
    )
    self.assertEqual(colored_text, original_text)
    self.assertEqual(text_formatting.decolored(colored_text), original_text)
    text_formatting.termcolor = termcolor


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Anthropic."""

import base64
import functools
import os
from typing import Annotated, Any, Literal

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import rest
import pyglove as pg


try:
  # pylint: disable=g-import-not-at-top
  from google import auth as google_auth
  from google.auth import credentials as credentials_lib
  from google.auth.transport import requests as auth_requests
  Credentials = credentials_lib.Credentials
  # pylint: enable=g-import-not-at-top
except ImportError:
  google_auth = None
  auth_requests = None
  credentials_lib = None
  Credentials = None  # pylint: disable=invalid-name


SUPPORTED_MODELS_AND_SETTINGS = {
    # See https://docs.anthropic.com/claude/docs/models-overview
    # Rate limits from https://docs.anthropic.com/claude/reference/rate-limits
    #     RPM/TPM for Claude-2.1, Claude-2.0, and Claude-Instant-1.2 estimated
    #     as RPM/TPM of the largest-available model (Claude-3-Opus).
    # Price in US dollars at https://www.anthropic.com/pricing
    # as of 2024-10-10.
    # Anthropic models hosted on VertexAI.
    'claude-3-5-sonnet-v2@20241022': pg.Dict(
        max_tokens=8192,
        rpm=1000,
        tpm=100000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.015,
    ),
    'claude-3-5-sonnet@20240620': pg.Dict(
        max_tokens=8192,
        rpm=1000,
        tpm=100000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.015,
    ),
    'claude-3-5-haiku@20241022': pg.Dict(
        max_tokens=8192,
        rpm=1000,
        tpm=100000,
        cost_per_1k_input_tokens=0.001,
        cost_per_1k_output_tokens=0.005,
    ),
    # Anthropic hosted models.
    'claude-3-5-sonnet-20241022': pg.Dict(
        max_tokens=8192,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.015,
    ),
    'claude-3-5-sonnet-20240620': pg.Dict(
        max_tokens=8192,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.015,
    ),
    'claude-3-5-haiku-20241022': pg.Dict(
        max_tokens=8192,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.001,
        cost_per_1k_output_tokens=0.005,
    ),
    'claude-3-opus-20240229': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.075,
    ),
    'claude-3-sonnet-20240229': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.015,
    ),
    'claude-3-haiku-20240307': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.00025,
        cost_per_1k_output_tokens=0.00125,
    ),
    'claude-2.1': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.008,
        cost_per_1k_output_tokens=0.024,
    ),
    'claude-2.0': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.008,
        cost_per_1k_output_tokens=0.024,
    ),
    'claude-instant-1.2': pg.Dict(
        max_tokens=4096,
        rpm=4000,
        tpm=400000,
        cost_per_1k_input_tokens=0.0008,
        cost_per_1k_output_tokens=0.0024,
    ),
}


@lf.use_init_args(['model'])
class Anthropic(rest.REST):
  """Anthropic LLMs (Claude) through REST APIs.

  See https://docs.anthropic.com/claude/reference/messages_post
  """

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ]

  multimodal: Annotated[bool, 'Whether this model has multimodal support.'] = (
      True
  )

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'ANTHROPIC_API_KEY'."
      ),
  ] = None

  api_endpoint: str = 'https://api.anthropic.com/v1/messages'

  api_version: Annotated[
      str,
      'Anthropic API version.'
  ] = '2023-06-01'

  def _on_bound(self):
    super()._on_bound()
    self._api_key = None

  def _initialize(self):
    api_key = self.api_key or os.environ.get('ANTHROPIC_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `ANTHROPIC_API_KEY` with your Anthropic API key.'
      )
    self._api_key = api_key

  @property
  def headers(self) -> dict[str, Any]:
    return {
        'x-api-key': self._api_key,
        'anthropic-version': self.api_version,
        'content-type': 'application/json',
        # TODO(yifenglu): Remove beta flag once the feature is fully supported.
        'anthropic-beta': 'pdfs-2024-09-25',
    }

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model

  @property
  def max_concurrency(self) -> int:
    rpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('rpm', 0)
    tpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_tokens', None
    )
    cost_per_1k_output_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_tokens', None
    )
    if cost_per_1k_output_tokens is None or cost_per_1k_input_tokens is None:
      return None
    return (
        cost_per_1k_input_tokens * num_input_tokens
        + cost_per_1k_output_tokens * num_output_tokens
    ) / 1000

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request = dict()
    request.update(self._request_args(sampling_options))
    request.update(
        dict(
            messages=[
                dict(role='user', content=self._content_from_message(prompt))
            ]
        )
    )
    return request

  def _request_args(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # Authropic requires `max_tokens` to be specified.
    max_tokens = (
        options.max_tokens
        or SUPPORTED_MODELS_AND_SETTINGS[self.model].max_tokens
    )
    args = dict(
        model=self.model,
        max_tokens=max_tokens,
        stream=False,
    )
    if options.stop:
      args['stop_sequences'] = options.stop
    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.top_k is not None:
      args['top_k'] = options.top_k
    if options.top_p is not None:
      args['top_p'] = options.top_p
    return args

  def _content_from_message(self, prompt: lf.Message) -> list[dict[str, Any]]:
    """Converts an message to Anthropic's content protocol (list of dicts)."""
    # Refer: https://docs.anthropic.com/claude/reference/messages-examples
    if self.multimodal:
      content = []
      for chunk in prompt.chunk():
        if isinstance(chunk, str):
          item = dict(type='text', text=chunk)
        elif isinstance(chunk, lf_modalities.Image):
          item = dict(
              type='image',
              source=dict(
                  type='base64',
                  media_type=chunk.mime_type,
                  data=base64.b64encode(chunk.to_bytes()).decode(),
              ),
          )
        elif isinstance(chunk, lf_modalities.PDF):
          item = dict(
              type='document',
              source=dict(
                  type='base64',
                  media_type=chunk.mime_type,
                  data=base64.b64encode(chunk.to_bytes()).decode(),
              ),
          )
        else:
          raise ValueError(f'Unsupported modality object: {chunk!r}.')
        content.append(item)
      return content
    else:
      return [dict(type='text', text=prompt.text)]

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    message = self._message_from_content(json['content'])
    input_tokens = json['usage']['input_tokens']
    output_tokens = json['usage']['output_tokens']
    return lf.LMSamplingResult(
        [lf.LMSample(message)],
        usage=lf.LMSamplingUsage(
            prompt_tokens=input_tokens,
            completion_tokens=output_tokens,
            total_tokens=input_tokens + output_tokens,
            estimated_cost=self.estimate_cost(
                num_input_tokens=input_tokens,
                num_output_tokens=output_tokens,
            ),
        ),
    )

  def _message_from_content(self, content: list[dict[str, Any]]) -> lf.Message:
    """Converts Anthropic's content protocol to message."""
    # Refer: https://docs.anthropic.com/claude/reference/messages-examples
    return lf.AIMessage.from_chunks(
        [x['text'] for x in content if x['type'] == 'text']
    )


class Claude3(Anthropic):
  """Base class for Claude 3 models. 200K input tokens and 4K output tokens."""
  multimodal = True


class Claude35Sonnet(Claude3):
  """A balance between between Opus and Haiku."""
  model = 'claude-3-5-sonnet-20241022'


class Claude35Sonnet20241022(Claude3):
  """A balance between between Opus and Haiku."""

  model = 'claude-3-5-sonnet-20241022'


class Claude35Sonnet20240620(Claude3):
  """A balance between between Opus and Haiku."""

  model = 'claude-3-5-sonnet-20240620'


class Claude3Opus(Claude3):
  """Anthropic's most powerful model."""

  model = 'claude-3-opus-20240229'


class Claude3Sonnet(Claude3):
  """A balance between between Opus and Haiku."""

  model = 'claude-3-sonnet-20240229'


class Claude3Haiku(Claude3):
  """Anthropic's most compact model."""

  model = 'claude-3-haiku-20240307'


class Claude2(Anthropic):
  """Predecessor to Claude 3 with 100K context window.."""
  model = 'claude-2.0'


class Claude21(Anthropic):
  """Updated Claude 2 model with improved accuracy and 200K context window."""
  model = 'claude-2.1'


class ClaudeInstant(Anthropic):
  """Cheapest small and fast model, 100K context window."""
  model = 'claude-instant-1.2'


#
# Authropic models on VertexAI.
#


class VertexAIAnthropic(Anthropic):
  """Anthropic models on VertexAI."""

  project: Annotated[
      str | None,
      'Google Cloud project ID.',
  ] = None

  location: Annotated[
      Literal['us-east5', 'europe-west1'],
      'GCP location with Anthropic models hosted.'
  ] = 'us-east5'

  credentials: Annotated[
      Credentials | None,    # pytype: disable=invalid-annotation
      (
          'Credentials to use. If None, the default credentials '
          'to the environment will be used.'
      ),
  ] = None

  api_version = 'vertex-2023-10-16'

  def _on_bound(self):
    super()._on_bound()
    if google_auth is None:
      raise ValueError(
          'Please install "langfun[llm-google-vertex]" to use Vertex AI models.'
      )
    self._project = None
    self._credentials = None

  def _initialize(self):
    project = self.project or os.environ.get('VERTEXAI_PROJECT', None)
    if not project:
      raise ValueError(
          'Please specify `project` during `__init__` or set environment '
          'variable `VERTEXAI_PROJECT` with your Vertex AI project ID.'
      )
    self._project = project
    credentials = self.credentials
    if credentials is None:
      # Use default credentials.
      credentials = google_auth.default(
          scopes=['https://www.googleapis.com/auth/cloud-platform']
      )
    self._credentials = credentials

  @functools.cached_property
  def _session(self):
    assert self._api_initialized
    assert self._credentials is not None
    assert auth_requests is not None
    s = auth_requests.AuthorizedSession(self._credentials)
    s.headers.update(self.headers or {})
    return s

  @property
  def headers(self):
    return {
        'Content-Type': 'application/json; charset=utf-8',
    }

  @property
  def api_endpoint(self) -> str:
    return (
        f'https://{self.location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self.location}/publishers/anthropic/'
        f'models/{self.model}:streamRawPredict'
    )

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ):
    request = super().request(prompt, sampling_options)
    request['anthropic_version'] = self.api_version
    del request['model']
    return request


class VertexAIClaude3_5_Sonnet_20241022(VertexAIAnthropic):  # pylint: disable=invalid-name
  """Anthropic's Claude 3.5 Sonnet model on VertexAI."""
  model = 'claude-3-5-sonnet-v2@20241022'


class VertexAIClaude3_5_Sonnet_20240620(VertexAIAnthropic):  # pylint: disable=invalid-name
  """Anthropic's Claude 3.5 Sonnet model on VertexAI."""
  model = 'claude-3-5-sonnet@20240620'


class VertexAIClaude3_5_Haiku_20241022(VertexAIAnthropic):  # pylint: disable=invalid-name
  """Anthropic's Claude 3.5 Haiku model on VertexAI."""
  model = 'claude-3-5-haiku@20241022'



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Groq."""

import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import rest
import pyglove as pg


SUPPORTED_MODELS_AND_SETTINGS = {
    # Refer https://console.groq.com/docs/models
    # Price in US dollars at https://groq.com/pricing/ as of 2024-10-10.
    'llama-3.2-3b-preview': pg.Dict(
        max_tokens=8192,
        max_concurrency=64,
        cost_per_1k_input_tokens=0.00006,
        cost_per_1k_output_tokens=0.00006,
    ),
    'llama-3.2-1b-preview': pg.Dict(
        max_tokens=8192,
        max_concurrency=64,
        cost_per_1k_input_tokens=0.00004,
        cost_per_1k_output_tokens=0.00004,
    ),
    'llama-3.1-70b-versatile': pg.Dict(
        max_tokens=8192,
        max_concurrency=16,
        cost_per_1k_input_tokens=0.00059,
        cost_per_1k_output_tokens=0.00079,
    ),
    'llama-3.1-8b-instant': pg.Dict(
        max_tokens=8192,
        max_concurrency=32,
        cost_per_1k_input_tokens=0.00005,
        cost_per_1k_output_tokens=0.00008,
    ),
    'llama3-70b-8192': pg.Dict(
        max_tokens=8192,
        max_concurrency=16,
        cost_per_1k_input_tokens=0.00059,
        cost_per_1k_output_tokens=0.00079,
    ),
    'llama3-8b-8192': pg.Dict(
        max_tokens=8192,
        max_concurrency=32,
        cost_per_1k_input_tokens=0.00005,
        cost_per_1k_output_tokens=0.00008,
    ),
    'llama2-70b-4096': pg.Dict(
        max_tokens=4096,
        max_concurrency=16,
    ),
    'mixtral-8x7b-32768': pg.Dict(
        max_tokens=32768,
        max_concurrency=16,
        cost_per_1k_input_tokens=0.00024,
        cost_per_1k_output_tokens=0.00024,
    ),
    'gemma2-9b-it': pg.Dict(
        max_tokens=8192,
        max_concurrency=32,
        cost_per_1k_input_tokens=0.0002,
        cost_per_1k_output_tokens=0.0002,
    ),
    'gemma-7b-it': pg.Dict(
        max_tokens=8192,
        max_concurrency=32,
        cost_per_1k_input_tokens=0.00007,
        cost_per_1k_output_tokens=0.00007,
    ),
    'whisper-large-v3': pg.Dict(
        max_tokens=8192,
        max_concurrency=16,
    ),
    'whisper-large-v3-turbo': pg.Dict(
        max_tokens=8192,
        max_concurrency=16,
    )
}


@lf.use_init_args(['model'])
class Groq(rest.REST):
  """Groq LLMs through REST APIs (OpenAI compatible).

  See https://platform.openai.com/docs/api-reference/chat
  """

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ]

  multimodal: Annotated[bool, 'Whether this model has multimodal support.'] = (
      False
  )

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'GROQ_API_KEY'."
      ),
  ] = None

  api_endpoint: str = 'https://api.groq.com/openai/v1/chat/completions'

  def _on_bound(self):
    super()._on_bound()
    self._api_key = None

  def _initialize(self):
    api_key = self.api_key or os.environ.get('GROQ_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `GROQ_API_KEY` with your Groq API key.'
      )
    self._api_key = api_key

  @property
  def headers(self) -> dict[str, Any]:
    return {
        'Authorization': f'Bearer {self._api_key}',
        'Content-Type': 'application/json',
    }

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model

  @property
  def max_concurrency(self) -> int:
    return SUPPORTED_MODELS_AND_SETTINGS[self.model].max_concurrency

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_tokens', None
    )
    cost_per_1k_output_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_tokens', None
    )
    if cost_per_1k_input_tokens is None or cost_per_1k_output_tokens is None:
      return None
    return (
        cost_per_1k_input_tokens * num_input_tokens
        + cost_per_1k_output_tokens * num_output_tokens
    ) / 1000

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request = dict()
    request.update(self._request_args(sampling_options))
    request.update(
        dict(
            messages=[
                dict(role='user', content=self._content_from_message(prompt))
            ]
        )
    )
    return request

  def _request_args(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # `logprobs` and `top_logprobs` flags are not supported on Groq yet.
    args = dict(
        model=self.model,
        n=options.n,
        stream=False,
    )

    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.max_tokens is not None:
      args['max_tokens'] = options.max_tokens
    if options.top_p is not None:
      args['top_p'] = options.top_p
    if options.stop:
      args['stop'] = options.stop
    return args

  def _content_from_message(self, prompt: lf.Message) -> list[dict[str, Any]]:
    """Converts an message to Groq's content protocol (list of dicts)."""
    # Refer: https://platform.openai.com/docs/api-reference/chat/create
    content = []
    for chunk in prompt.chunk():
      if isinstance(chunk, str):
        item = dict(type='text', text=chunk)
      elif (
          self.multimodal
          and isinstance(chunk, lf_modalities.Image)
          and chunk.uri
      ):
        # NOTE(daiyip): Groq only support image URL.
        item = dict(type='image_url', image_url=chunk.uri)
      else:
        raise ValueError(f'Unsupported modality object: {chunk!r}.')
      content.append(item)
    return content

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    samples = [
        lf.LMSample(self._message_from_choice(choice), score=0.0)
        for choice in json['choices']
    ]
    usage = json['usage']
    return lf.LMSamplingResult(
        samples,
        usage=lf.LMSamplingUsage(
            prompt_tokens=usage['prompt_tokens'],
            completion_tokens=usage['completion_tokens'],
            total_tokens=usage['total_tokens'],
            estimated_cost=self.estimate_cost(
                num_input_tokens=usage['prompt_tokens'],
                num_output_tokens=usage['completion_tokens'],
            ),
        ),
    )

  def _message_from_choice(self, choice: dict[str, Any]) -> lf.Message:
    """Converts Groq's content protocol to message."""
    # Refer: https://platform.openai.com/docs/api-reference/chat/create
    content = choice['message']['content']
    if isinstance(content, str):
      return lf.AIMessage(content)
    return lf.AIMessage.from_chunks(
        [x['text'] for x in content if x['type'] == 'text']
    )


class GroqLlama3_2_3B(Groq):  # pylint: disable=invalid-name
  """Llama3.2-3B with 8K context window.

  See: https://huggingface.co/meta-llama/Llama-3.2-3B
  """

  model = 'llama-3.2-3b-preview'


class GroqLlama3_2_1B(Groq):  # pylint: disable=invalid-name
  """Llama3.2-1B with 8K context window.

  See: https://huggingface.co/meta-llama/Llama-3.2-1B
  """

  model = 'llama-3.2-3b-preview'


class GroqLlama3_8B(Groq):  # pylint: disable=invalid-name
  """Llama3-8B with 8K context window.

  See: https://huggingface.co/meta-llama/Meta-Llama-3-8B
  """

  model = 'llama3-8b-8192'


class GroqLlama3_1_70B(Groq):  # pylint: disable=invalid-name
  """Llama3.1-70B with 8K context window.

  See: https://github.com/meta-llama/llama-models/blob/main/models/llama3_1/MODEL_CARD.md   # pylint: disable=line-too-long
  """

  model = 'llama-3.1-70b-versatile'


class GroqLlama3_1_8B(Groq):  # pylint: disable=invalid-name
  """Llama3.1-8B with 8K context window.

  See: https://github.com/meta-llama/llama-models/blob/main/models/llama3_1/MODEL_CARD.md   # pylint: disable=line-too-long
  """

  model = 'llama-3.1-8b-instant'


class GroqLlama3_70B(Groq):  # pylint: disable=invalid-name
  """Llama3-70B with 8K context window.

  See: https://huggingface.co/meta-llama/Meta-Llama-3-70B
  """

  model = 'llama3-70b-8192'


class GroqLlama2_70B(Groq):  # pylint: disable=invalid-name
  """Llama2-70B with 4K context window.

  See: https://huggingface.co/meta-llama/Llama-2-70b
  """

  model = 'llama2-70b-4096'


class GroqMistral_8x7B(Groq):  # pylint: disable=invalid-name
  """Mixtral 8x7B with 32K context window.

  See: https://huggingface.co/meta-llama/Llama-2-70b
  """

  model = 'mixtral-8x7b-32768'


class GroqGemma2_9B_IT(Groq):  # pylint: disable=invalid-name
  """Gemma2 9B with 8K context window.

  See: https://huggingface.co/google/gemma-2-9b-it
  """

  model = 'gemma2-9b-it'


class GroqGemma_7B_IT(Groq):  # pylint: disable=invalid-name
  """Gemma 7B with 8K context window.

  See: https://huggingface.co/google/gemma-1.1-7b-it
  """

  model = 'gemma-7b-it'


class GroqWhisper_Large_v3(Groq):  # pylint: disable=invalid-name
  """Whisper Large V3 with 8K context window.

  See: https://huggingface.co/openai/whisper-large-v3
  """

  model = 'whisper-large-v3'


class GroqWhisper_Large_v3Turbo(Groq):  # pylint: disable=invalid-name
  """Whisper Large V3 Turbo with 8K context window.

  See: https://huggingface.co/openai/whisper-large-v3-turbo
  """

  model = 'whisper-large-v3-turbo'



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for agentic actions."""

import abc
from typing import Annotated, Any, Optional, Union
import langfun.core as lf
import pyglove as pg


class Action(pg.Object):
  """Base class for agent actions."""

  def _on_bound(self):
    super()._on_bound()
    self._result = None

  @property
  def result(self) -> Any:
    """Returns the result of the action."""
    return self._result

  def __call__(
      self, session: Optional['Session'] = None, **kwargs) -> Any:
    """Executes the action."""
    session = session or Session()
    try:
      session.begin(self)
      self._result = self.call(session=session, **kwargs)
      return self._result
    finally:
      session.end(self)

  @abc.abstractmethod
  def call(self, session: 'Session', **kwargs) -> Any:
    """Subclasses to implement."""


class ActionInvocation(pg.Object, pg.views.html.HtmlTreeView.Extension):
  """A class for capturing the invocation of an action."""
  action: Action
  result: Any = None
  execution: Annotated[
      list[Union['ActionInvocation', lf.logging.LogEntry]],
      'Execution execution.'
  ] = []

  # Allow symbolic assignment without `rebind`.
  allow_symbolic_assignment = True

  @property
  def logs(self) -> list[lf.logging.LogEntry]:
    """Returns logs from execution sequence."""
    return [v for v in self.execution if isinstance(v, lf.logging.LogEntry)]

  @property
  def child_invocations(self) -> list['ActionInvocation']:
    """Returns child action invocations."""
    return [v for v in self.execution if isinstance(v, ActionInvocation)]

  def _html_tree_view_summary(
      self, *, view: pg.views.html.HtmlTreeView, **kwargs
  ):
    if isinstance(self.action, RootAction):
      return None
    kwargs.pop('title')
    return view.summary(
        self,
        title=view.render(
            self.action, name='action', collapse_level=0,
            css_classes='invocation-title',
        ),
        **kwargs
    )

  def _html_tree_view_content(
      self,
      *,
      root_path: pg.KeyPath | None = None,
      collapse_level: int | None = None,
      view: pg.views.html.HtmlTreeView,
      **kwargs
  ):
    prepare_phase = []
    current_phase = prepare_phase
    action_phases = []
    for item in self.execution:
      if isinstance(item, ActionInvocation):
        current_phase = []
        action_phases.append(current_phase)
      current_phase.append(item)

    def _render_phase(
        phase: list[ActionInvocation | lf.logging.LogEntry]
    ) -> pg.Html.WritableTypes:
      return pg.Html.element(
          'div',
          [
              view.render(item) for item in phase
          ]
      )

    def _render_action_phases(
        phases: list[list[ActionInvocation | lf.logging.LogEntry]]
    ) -> pg.Html.WritableTypes:
      if len(phases) == 1:
        return _render_phase(phases[0])
      return pg.views.html.controls.TabControl(
          [
              pg.views.html.controls.Tab(
                  label=f'Step {i + 1}',
                  content=_render_phase(phase),
              )
              for i, phase in enumerate(phases)
          ],
      )

    result_name = 'final_result' if isinstance(
        self.action, RootAction) else 'result'
    return pg.Html.element(
        'div',
        [
            view.render(
                self.result,
                name=result_name,
                css_classes=[
                    f'invocation-{result_name}'.replace('_', '-')
                ]
            ),
            _render_phase(prepare_phase) if prepare_phase else None,
            _render_action_phases(action_phases)
        ]
    )

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        details.invocation-title {
          display: inline-block;
          background-color: #b1f0ff;
          border: 1px solid white;
        }
        details.invocation-result {
          border: 1px solid #eee;
        }
        details.invocation-final-result {
          border: 1px solid #eee;
          background-color: #fef78f;
        }
        """
    ]


class RootAction(Action):
  """A placeholder action for the root of the action tree."""

  def call(self, session: 'Session', **kwargs) -> Any:
    raise NotImplementedError('Shall not be called.')


class Session(pg.Object):
  """Session for performing an agentic task."""

  root_invocation: ActionInvocation = ActionInvocation(RootAction())

  def _on_bound(self):
    super()._on_bound()
    self._invocation_stack = [self.root_invocation]

  @property
  def final_result(self) -> Any:
    """Returns the final result of the session."""
    return self.root_invocation.result

  @property
  def current_invocation(self) -> ActionInvocation:
    """Returns the current invocation."""
    assert self._invocation_stack
    return self._invocation_stack[-1]

  def begin(self, action: Action):
    """Signal the beginning of the execution of an action."""
    new_invocation = ActionInvocation(pg.maybe_ref(action))
    with pg.notify_on_change(False):
      self.current_invocation.execution.append(new_invocation)
    self._invocation_stack.append(new_invocation)

  def end(self, action: Action):
    """Signal the end of the execution of an action."""
    assert self._invocation_stack
    invocation = self._invocation_stack.pop(-1)
    invocation.rebind(
        result=action.result, skip_notification=True, raise_on_no_change=False
    )
    assert invocation.action is action, (invocation.action, action)
    assert self._invocation_stack, self._invocation_stack

    if len(self._invocation_stack) == 1:
      self.root_invocation.rebind(
          result=invocation.result,
          skip_notification=True,
          raise_on_no_change=False
      )

  def _log(self, level: lf.logging.LogLevel, message: str, **kwargs):
    with pg.notify_on_change(False):
      self.current_invocation.execution.append(
          lf.logging.log(
              level, message, indent=len(self._invocation_stack) - 1, **kwargs
          )
      )

  def debug(self, message: str, **kwargs):
    """Logs a debug message to the session."""
    self._log('debug', message, **kwargs)

  def info(self, message: str, **kwargs):
    """Logs an info message to the session."""
    self._log('info', message, **kwargs)

  def warning(self, message: str, **kwargs):
    """Logs a warning message to the session."""
    self._log('warning', message, **kwargs)

  def error(self, message: str, **kwargs):
    """Logs an error message to the session."""
    self._log('error', message, **kwargs)

  def fatal(self, message: str, **kwargs):
    """Logs a fatal message to the session."""
    self._log('fatal', message, **kwargs)

  def as_message(self) -> lf.AIMessage:
    """Returns the session as a message."""
    return lf.AIMessage(
        'Agentic task session.',
        result=self.root_invocation
    )



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for implementing _repr_xxx_ methods."""

import collections
import contextlib
import html
import io
from typing import Any, Callable, Iterator

from langfun.core import component
import pyglove as pg


class Html(pg.Object):
  """A HTML adapter for rendering."""
  content: str

  def _repr_html_(self) -> str:
    return self.content


@contextlib.contextmanager
def share_parts() -> Iterator[dict[str, int]]:
  """Context manager for defining the context (scope) of shared content.

  Under the context manager, call to `lf.write_shared` with the same content
  will be written only once. This is useful for writing shared content such as
  shared style and script sections in the HTML.

  Example:
    ```
    class Foo(pg.Object):
      def _repr_html_(self) -> str:
        s = io.StringIO()
        lf.repr_utils.write_shared_part(s, '<style>..</style>')
        lf.repr_utils.write_shared_part(s, '<script>..</script>')
        return s.getvalue()

    with lf.repr_utils.share_parts() as share_parts:
      # The <style> and <script> section will be written only once.
      lf.console.display(Foo())
      lf.console.display(Foo())

    # Assert that the shared content is attempted to be written twice.
    assert share_parts['<style>..</style>'] == 2
    ```

  Yields:
    A dictionary mapping the shared content to the number of times it is
    attempted to be written.
  """
  context = component.context_value(
      '__shared_parts__', collections.defaultdict(int)
  )
  with component.context(__shared_parts__=context):
    try:
      yield context
    finally:
      pass


def write_maybe_shared(s: io.StringIO, content: str) -> bool:
  """Writes a maybe shared part to an string stream.

  Args:
    s: The string stream to write to.
    content: A maybe shared content to write.

  Returns:
    True if the content is written to the string. False if the content is
    already written under the same share context.
  """
  context = component.context_value('__shared_parts__', None)
  if context is None:
    s.write(content)
    return True
  written = content in context
  if not written:
    s.write(content)
  context[content] += 1
  return not written


def html_repr(
    value: dict[str, Any],
    item_color: Callable[
        [str, str],
        tuple[
            str | None,   # Label text color
            str | None,   # Label background color
            str | None,   # Value text color
            str | None,   # Value background color
        ]
    ] | None = None    # pylint: disable=bad-whitespace
) -> str:
  """Writes a list of key-value pairs to an string stream.

  Args:
    value: A value to be rendered in HTML.
    item_color: A function that takes the key and value and returns a tuple
      of four strings, the text color and background color of the label and
      value respectively. If None, a default color scheme will be used.

  Returns:
    The HTML representation of the value.
  """
  s = io.StringIO()
  s.write('<div style="padding-left: 20px; margin-top: 10px">')
  s.write('<table style="border-top: 1px solid #EEEEEE;">')
  item_color = item_color or (lambda k, v: (None, '#F1C40F', None, None))

  def maybe_html_format(v: Any, root_indent: int) -> str | None:
    del root_indent
    if hasattr(v, '_repr_html_'):
      return v._repr_html_()  # pylint: disable=protected-access
    # Fall back to the default format.
    return None

  with (pg.str_format(custom_format=maybe_html_format),
        pg.repr_format(custom_format=maybe_html_format)):
    for k, v in pg.object_utils.flatten(value).items():
      if isinstance(v, pg.Ref):
        v = v.value
      if hasattr(v, '_repr_html_'):
        cs = v._repr_html_()  # pylint: disable=protected-access
      else:
        cs = html.escape(v) if isinstance(v, str) else escape_quoted(str(v))
        cs = f'<span style="white-space: pre-wrap">{cs}</span>'

      key_color, key_bg_color, value_color, value_bg_color = item_color(k, v)
      key_span = html_round_text(
          k,
          text_color=key_color,
          background_color=key_bg_color,
          margin_bottom='0px'
      )
      value_color_style = f'color: {value_color};' if value_color else ''
      value_bg_color_style = (
          f'background-color: {value_bg_color};' if value_bg_color else ''
      )
      s.write(
          '<tr>'
          '<td style="padding: 5px; vertical-align: top; '
          f'border-bottom: 1px solid #EEEEEE">{key_span}</td>'
          '<td style="padding: 15px 5px 5px 5px; vertical-align: top; '
          'border-bottom: 1px solid #EEEEEE;'
          f'{value_color_style}{value_bg_color_style}">{cs}</td></tr>'
      )
    s.write('</table></div>')
    return s.getvalue()


def html_round_text(
    text: str,
    *,
    text_color: str = 'black',
    background_color: str = '#EEEEEE',
    display: str = 'inline-block',
    margin_top: str = '5px',
    margin_bottom: str = '5px',
    whitespace: str = 'pre-wrap') -> str:
  """Renders a HTML span with rounded corners."""
  color_style = f'color: {text_color};' if text_color else ''
  bg_color_style = (
      f'background-color: {background_color};' if background_color else ''
  )
  return (
      f'<span style="{color_style}{bg_color_style}'
      f'display:{display}; border-radius:10px; padding:5px; '
      f'margin-top: {margin_top}; margin-bottom: {margin_bottom}; '
      f'white-space: {whitespace}">{text}</span>'
  )


def escape_quoted(s: str):
  """Escape quoted parts within a string."""
  r = io.StringIO()
  quote_char = None
  quote_start = -1
  for i, c in enumerate(s):
    if c in ('\'', '"'):
      if quote_char is None:
        quote_char = c
        quote_start = i
      elif quote_char == c:
        r.write(c)
        r.write(html.escape(s[quote_start + 1:i]))
        r.write(c)
        quote_char = None
    elif quote_char is None:
      r.write(c)
  return r.getvalue()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Gemini models."""

import os
import unittest
from unittest import mock

from google import generativeai as genai
import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import google_genai
import pyglove as pg


example_image = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


def mock_get_model(model_name, *args, **kwargs):
  del args, kwargs
  if 'gemini' in model_name:
    method = 'generateContent'
  elif 'chat' in model_name:
    method = 'generateMessage'
  else:
    method = 'generateText'
  return pg.Dict(supported_generation_methods=[method])


def mock_generate_text(*, model, prompt, **kwargs):
  return pg.Dict(
      candidates=[pg.Dict(output=f'{prompt} to {model} with {kwargs}')]
  )


def mock_chat(*, model, messages, **kwargs):
  return pg.Dict(
      candidates=[pg.Dict(content=f'{messages} to {model} with {kwargs}')]
  )


def mock_generate_content(content, generation_config, **kwargs):
  del kwargs
  c = generation_config
  return genai.types.GenerateContentResponse(
      done=True,
      iterator=None,
      chunks=[],
      result=pg.Dict(
          prompt_feedback=pg.Dict(block_reason=None),
          candidates=[
              pg.Dict(
                  content=pg.Dict(
                      parts=[
                          pg.Dict(
                              text=(
                                  f'This is a response to {content[0]} with '
                                  f'n={c.candidate_count}, '
                                  f'temperature={c.temperature}, '
                                  f'top_p={c.top_p}, '
                                  f'top_k={c.top_k}, '
                                  f'max_tokens={c.max_output_tokens}, '
                                  f'stop={c.stop_sequences}.'
                              )
                          )
                      ]
                  ),
              ),
          ],
      ),
  )


class GenAITest(unittest.TestCase):
  """Tests for Google GenAI model."""

  def test_content_from_message_text_only(self):
    text = 'This is a beautiful day'
    model = google_genai.GeminiPro()
    chunks = model._content_from_message(lf.UserMessage(text))
    self.assertEqual(chunks, [text])

  def test_content_from_message_mm(self):
    message = lf.UserMessage(
        'This is an <<[[image]]>>, what is it?',
        image=lf_modalities.Image.from_bytes(example_image),
    )

    # Non-multimodal model.
    with self.assertRaisesRegex(lf.ModalityError, 'Unsupported modality'):
      google_genai.GeminiPro()._content_from_message(message)

    model = google_genai.GeminiProVision()
    chunks = model._content_from_message(message)
    self.maxDiff = None
    self.assertEqual(
        chunks,
        [
            'This is an',
            genai.types.BlobDict(mime_type='image/png', data=example_image),
            ', what is it?',
        ],
    )

  def test_response_to_result_text_only(self):
    response = genai.types.GenerateContentResponse(
        done=True,
        iterator=None,
        chunks=[],
        result=pg.Dict(
            prompt_feedback=pg.Dict(block_reason=None),
            candidates=[
                pg.Dict(
                    content=pg.Dict(
                        parts=[pg.Dict(text='This is response 1.')]
                    ),
                ),
                pg.Dict(
                    content=pg.Dict(parts=[pg.Dict(text='This is response 2.')])
                ),
            ],
        ),
    )
    model = google_genai.GeminiProVision()
    result = model._response_to_result(response)
    self.assertEqual(
        result,
        lf.LMSamplingResult([
            lf.LMSample(lf.AIMessage('This is response 1.'), score=0.0),
            lf.LMSample(lf.AIMessage('This is response 2.'), score=0.0),
        ]),
    )

  def test_model_hub(self):
    orig_get_model = genai.get_model
    genai.get_model = mock_get_model

    model = google_genai._GOOGLE_GENAI_MODEL_HUB.get('gemini-pro')
    self.assertIsNotNone(model)
    self.assertIs(google_genai._GOOGLE_GENAI_MODEL_HUB.get('gemini-pro'), model)

    genai.get_model = orig_get_model

  def test_api_key_check(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      _ = google_genai.GeminiPro()._api_initialized

    self.assertTrue(google_genai.GeminiPro(api_key='abc')._api_initialized)
    os.environ['GOOGLE_API_KEY'] = 'abc'
    self.assertTrue(google_genai.GeminiPro()._api_initialized)
    del os.environ['GOOGLE_API_KEY']

  def test_call(self):
    with mock.patch(
        'google.generativeai.GenerativeModel.generate_content',
    ) as mock_generate:
      orig_get_model = genai.get_model
      genai.get_model = mock_get_model
      mock_generate.side_effect = mock_generate_content

      lm = google_genai.GeminiPro(api_key='test_key')
      self.maxDiff = None
      self.assertEqual(
          lm('hello', temperature=2.0, top_k=20, max_tokens=1024).text,
          (
              'This is a response to hello with n=1, temperature=2.0, '
              'top_p=None, top_k=20, max_tokens=1024, stop=None.'
          ),
      )
      genai.get_model = orig_get_model

  def test_call_with_legacy_completion_model(self):
    orig_get_model = genai.get_model
    genai.get_model = mock_get_model
    orig_generate_text = getattr(genai, 'generate_text', None)
    if orig_generate_text is not None:
      genai.generate_text = mock_generate_text

      lm = google_genai.Palm2(api_key='test_key')
      self.maxDiff = None
      self.assertEqual(
          lm('hello', temperature=2.0, top_k=20).text,
          (
              "hello to models/text-bison-001 with {'temperature': 2.0, "
              "'top_k': 20, 'top_p': None, 'candidate_count': 1, "
              "'max_output_tokens': None, 'stop_sequences': None}"
          ),
      )
      genai.generate_text = orig_generate_text
    genai.get_model = orig_get_model

  def test_call_with_legacy_chat_model(self):
    orig_get_model = genai.get_model
    genai.get_model = mock_get_model
    orig_chat = getattr(genai, 'chat', None)
    if orig_chat is not None:
      genai.chat = mock_chat

      lm = google_genai.Palm2_IT(api_key='test_key')
      self.maxDiff = None
      self.assertEqual(
          lm('hello', temperature=2.0, top_k=20).text,
          (
              "hello to models/chat-bison-001 with {'temperature': 2.0, "
              "'top_k': 20, 'top_p': None, 'candidate_count': 1}"
          ),
      )
      genai.chat = orig_chat
    genai.get_model = orig_get_model


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from llama.cpp."""

from typing import Any

import langfun.core as lf
from langfun.core.llms import rest
import pyglove as pg


class LlamaCppRemote(rest.REST):
  """The remote LLaMA C++ model.

  The Remote LLaMA C++ models can be launched via
  https://github.com/ggerganov/llama.cpp/tree/master/examples/server
  """

  @pg.explicit_method_override
  def __init__(self, url: str, model: str | None = None, **kwargs):
    super().__init__(api_endpoint=f'{url}/completion', model=model, **kwargs)

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'LLaMAC++({self.model or ""})'

  def request(
      self, prompt: lf.Message, sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request = dict()
    request.update(self._request_args(sampling_options))
    # NOTE(daiyip): multi-modal is current not supported.
    request['prompt'] = prompt.text
    return request

  def _request_args(self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    args = dict(
        n_predict=options.max_tokens or 1024,
        top_k=options.top_k or 50,
        top_p=options.top_p or 0.95,
    )
    if options.temperature is not None:
      args['temperature'] = options.temperature
    return args

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    return lf.LMSamplingResult(
        [lf.LMSample(item['content'], score=0.0) for item in json['items']]
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    request = self.request(prompt, self.sampling_options)

    def _sample_one_example(request):
      response = self._session.post(
          self.api_endpoint,
          json=request,
          timeout=self.timeout,
      )
      if response.status_code == 200:
        return response.json()
      else:
        error_cls = self._error_cls_from_status(response.status_code)
        raise error_cls(f'{response.status_code}: {response.content}')

    items = self._parallel_execute_with_currency_control(
        _sample_one_example,
        [request] * (self.sampling_options.n or 1),
    )
    return self.result(dict(items=items))
